# Copyright (C) 2010-2011
# Author: Kurakin Roman, <rik@inse.ru>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

use Globus::GRAM::Error;
use Globus::GRAM::JobState;
use Globus::GRAM::JobManager;
use Globus::Core::Paths;
use Globus::Core::Config;

use IO::File;
use IPC::Open3;
use Config;

# NOTE: This package name must match the name of the .pm file!!
package Globus::GRAM::JobManager::slurm;

@ISA = qw(Globus::GRAM::JobManager);

my ($mpirun, $sbatch, $srun, $scontrol, $scancel, $salloc, $cpu_per_node);

BEGIN
{
    my $config = new Globus::Core::Config(
            '${sysconfdir}/globus/globus-slurm.conf');

    $mpirun = $config->get_attribute('mpirun') || 'no';
    if ($mpirun ne 'no' && ! -x $mpirun)
    {
        $mpirun = 'no';
    }
    $srun = $config->get_attribute('srun') || 'no';
    $sbatch = $config->get_attribute('sbatch') || 'no';
    $scancel = $config->get_attribute('scancel') || 'no';
    $scontrol = $config->get_attribute('scontrol') || 'no';
    $salloc = $config->get_attribute('salloc') || 'no';
    $slurm_path= $config->get_attribute('slurm_path') || 'no';
    $mpi_flavor= $config->get_attribute('mpi_flavor') || 'no';
    $mpi_path= $config->get_attribute('mpi_path') || 'no';
    $cpu_per_node   = 1;
    $softenv_dir    = '';
    $soft_msc       = "$softenv_dir/bin/soft-msc";
    $softenv_load   = "$softenv_dir/etc/softenv-load.sh";
}

sub myceil ($)
{
    my $x = shift;
    ( abs($x-int($x)) < 1E-12 ) ? int($x) : int($x < 0 ? $x : $x+1.0);
}

sub submit
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $status;
    my $slurm_job_script;
    my $slurm_job_script_name;
    my $slurm_sbatch_err_name ;
    my $errfile = '';
    my $job_id;
    my $rsh_env;
    my @arguments;
    my $email_when = '';
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";
    my %library_vars;
    my $max_time = 0;
    my $max_cpu_time = 0;
    my $max_wall_time = 0;
    my $count = 1;
    my $host_count = 0;

    $self->log("Entering slurm submit");

    # Reject jobs that want streaming, if so configured
    if ($description->streamingrequested() &&
	$description->streamingdisabled() ) {

	$self->log("Streaming is not allowed.");
	return Globus::GRAM::Error::OPENING_STDOUT;
    }

    # check jobtype
    if (defined($description->jobtype())) {
        if ($description->jobtype !~ /^(mpi|single|multiple)$/) {
            return Globus::GRAM::Error::JOBTYPE_NOT_SUPPORTED;
        }
    }
    if ($description->directory eq '') {
        return Globus::GRAM::Error::RSL_DIRECTORY();
    }
    if ($description->directory() =~ m|^[^/]|) {
        $description->add("directory",
                $ENV{HOME} . '/' . $description->directory());
    }
    chdir $description->directory() or
        return Globus::GRAM::Error::BAD_DIRECTORY();

    $self->nfssync( $description->executable() )
        unless $description->executable() eq '';
    $self->nfssync( $description->stdin() )
        unless $description->stdin() eq '';
    if ($description->executable eq '') {
        return Globus::GRAM::Error::RSL_EXECUTABLE();
    } elsif (! -f $description->executable()) {
        return Globus::GRAM::Error::EXECUTABLE_NOT_FOUND();
    } elsif (! -x $description->executable()) {
        return Globus::GRAM::Error::EXECUTABLE_PERMISSIONS();
    } elsif ($description->stdin() eq '') {
        return Globus::GRAM::Error::RSL_STDIN;
    } elsif (! -r $description->stdin()) {
       return Globus::GRAM::Error::STDIN_NOT_FOUND();
    }

    $max_time = $description->max_time() if (defined($description->max_time()));
    $max_cpu_time = $description->max_cpu_time() if (defined($description->max_cpu_time()));
    $max_wall_time = $description->max_wall_time() if (defined($description->max_wall_time()));

    $self->log("Determining job max time cpu from job description");
    if ($max_cpu_time == 0 && $max_time != 0) {
        $max_cpu_time = $max_time;
        $self->log("   using max_time as max_cpu_time");
    }
    $self->log("Determining job max wall time limit from job description");
    if ($max_wall_time == 0 && $max_time != 0) {
        $max_wall_time = $max_time;
        $self->log("   using max_time as max_wall_time");
    }

    $self->log('Building job script');

    $slurm_job_script_name = $self->job_dir() . '/scheduler_slurm_job_script';

    local(*JOB);
    open( JOB, '>' . $slurm_job_script_name );
    print JOB<<"EOF";
#! /bin/sh
# SLURM batch job script built by Globus job manager
#
EOF

    if ($description->name() ne '') {
        print JOB '#SBATCH -J ', $description->name(), "\n";
    }
#    # Only local users are supported.
#    if ($description->email_address() ne '') {
#        print JOB '#SBATCH --mail-user=', $description->email_address(), "\n";
#    }
    if ($description->emailonabort() eq 'yes') {
        $email_when .= 'a';
    }
    if ($description->emailonexecution() eq 'yes') {
        $email_when .= 'b';
    }
    if ($description->emailontermination() eq 'yes') {
        $email_when .= 'e';
    }
    if ($email_when eq 'abe') {
        $email_when = 'ALL';
    } elsif ($email_when eq 'a') {
        $email_when = 'FAIL';
    } elsif ($email_when eq 'b') {
        $email_when = 'BEGIN';
    } elsif ($email_when eq 'e') {
        $email_when = 'END';
    }
#    print JOB "#SBATCH --mail-type=$email_when\n" unless $email_when eq '';

    if ($description->queue() ne '') {
        print JOB '#SBATCH -p ', $description->queue(), "\n";
    }
#    if ($description->project() ne '') {
#        print JOB '#SBATCH -A ', $description->project(), "\n";
#    }

    # sbatch does not support cpu-time limit, use only wall-time.
    # The zero value is accepted as no limit.
    print JOB '#SBATCH -t ', $max_wall_time, "\n";

    print JOB '#SBATCH -o ', $description->stdout(), "\n";
    print JOB '#SBATCH -e ', $description->stderr(), "\n";
    print JOB '#SBATCH -i ', $description->stdin(), "\n";

    $count = $description->count() if (defined($description->count()));
    $host_count = $description->host_count() if (defined($description->host_count()));
    
    if ($host_count) {
        print JOB '#SBATCH -N ', $host_count, "\n";
        print JOB '#SBATCH --ntasks-per-node=', $count, "\n";
        if ($description->max_memory() != 0) {
            print JOB '#SBATCH --mem=', $description->max_memory(), "\n";
        }
    } else {
        print JOB '#SBATCH -n ', $count, "\n";
        if ($description->max_memory() != 0) {
            print JOB '#SBATCH --mem-per-cpu=', $description->max_memory(), "\n";
        }
    }

    ### SoftEnv extension ###
    if ($softenv_dir ne '')
    {
        $self->setup_softenv(
            $self->job_dir() . '/slurm_softenv_job_script',
            $soft_msc,
            $softenv_load,
            *JOB);
    }
    #########################

    $rsh_env = '';

    $library_vars{LD_LIBRARY_PATH} = 0;
    if ($^O eq 'irix') {
        $library_vars{LD_LIBRARYN32_PATH} = 0;
        $library_vars{LD_LIBRARY64_PATH} = 0;
    }

    foreach my $tuple ($description->environment()) {
        if (!ref($tuple) || scalar(@$tuple) != 2) {
            return Globus::GRAM::Error::RSL_ENVIRONMENT();
        }
        if (exists($library_vars{$tuple->[0]})) {
            $tuple->[1] .= ":$library_string";
            $library_vars{$tuple->[0]} = 1;
        }

        push(@new_env, $tuple->[0] . '="' . $tuple->[1] . '"');

        $tuple->[0] =~ s/\\/\\\\/g;
        $tuple->[0] =~ s/\$/\\\$/g;
        $tuple->[0] =~ s/"/\\\"/g; #"
        $tuple->[0] =~ s/`/\\\`/g; #`

        $tuple->[1] =~ s/\\/\\\\/g;
        $tuple->[1] =~ s/\$/\\\$/g;
        $tuple->[1] =~ s/"/\\\"/g; #"
        $tuple->[1] =~ s/`/\\\`/g; #`

        $rsh_env .= $tuple->[0] . '="' . $tuple->[1] . "\";\n"
                 .  'export ' . $tuple->[0] . ";\n";
    }
    foreach (keys %library_vars) {
        if ($library_vars{$_} == 0) {
            my $library_path = $description->library_path();
            push(@new_env, $_ . '=' . $library_path);
            $rsh_env .= "$_=$library_path;\n"
                      . "export $_;\n";
        }
    }

   

    print JOB "$rsh_env";

    print JOB "\n#Change to directory requested by user\n";
    print JOB 'cd ' . $description->directory() . "\n";

    @arguments = $description->arguments();

    foreach (@arguments) {
        if (ref($_)) {
            return Globus::GRAM::Error::RSL_ARGUMENTS;
        }
    }
    if ($#arguments >= 0) {
        foreach (@arguments) {
            $self->log("Transforming argument \"$_\"\n");
            $_ =~ s/\\/\\\\/g;
            $_ =~ s/\$/\\\$/g;
            $_ =~ s/"/\\\"/g; #"
            $_ =~ s/`/\\\`/g; #`
            $self->log("Transformed to \"$_\"\n");

            $args .= '"' . $_ . '" ';
        }
    } else {
        $args = '';
    }

    if ($description->jobtype() eq 'mpi') {
	if ($mpi_flavor eq 'openmpi') {
	    print JOB "export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$slurm_path/lib:$mpi_path/lib\n";
	    print JOB "export PATH=\$PATH:$slurm_path/bin:$mpi_path/bin\n";
	}
   }

    if ($description->executable() =~ m|^[^/]|) {
        $description->add('executable', './' . $description->executable());
    }
    if ($description->jobtype() eq 'multiple') {
        print JOB "$srun ", $description->executable(), " $args\n";
    } elsif ($description->jobtype() eq 'mpi') {
	if ( $mpi_flavor eq 'openmpi'){
            print JOB "$salloc $mpirun ", $description->executable(), " $args\n";}
	elsif ($mpi_flavor eq 'mpich2'){
	    print JOB "$srun $mpirun ", $description->executable(), "$args\n";}
    } else {
        print JOB $description->executable(), " $args\n";
    }
    close(JOB);

    $slurm_sbatch_err_name = $self->job_dir() . '/scheduler_slurm_submit_stderr';
    $errfile = "2>$slurm_sbatch_err_name";

    $self->nfssync( $slurm_job_script_name );
    $self->nfssync( $slurm_sbatch_err_name );
    $self->nfssync( $description->stdout, 1 );
    $self->nfssync( $description->stderr, 1 );
    $self->log("submitting job -- $sbatch $slurm_job_script_name $errfile");
    $sub_res = `cd $job_dir; $sbatch $slurm_job_script_name $errfile`;
    $job_id = (grep(/^Submitted batch job \d+$/, split(/\n/, $sub_res)))[0];

    if ($? == 0 && defined $job_id) {
        $job_id =~ s/^Submitted batch job (\d+)$/\1/;
        $self->log("job submission successful, setting state to PENDING");
        return {JOB_ID => $job_id,
                JOB_STATE => Globus::GRAM::JobState::PENDING };
    }

    if ($? == 0 && not defined $job_id) {
        local(*ERR);
        local $/;
        $self->log("failed to findout job id, sbatch returned $sub_res");

        open(ERR, ">" . $description->stderr());
        print ERR "Failed to findout job id, sbatch returned $sub_res\n";
        print ERR "If you are sure this is target resource failure, ";
        print ERR "please, contact\n";
        print ERR "resource administrator.\n";
        close(ERR);

        $sub_res =~ s/\n/\\n/g;
        $self->respond({GT3_FAILURE_MESSAGE => $sub_res});
    } else {
        local(*ERR);
        local $/;

        open(ERR, "<$slurm_sbatch_err_name");
        my $stderr = <ERR>;
        close(ERR);

        $self->log("sbatch returned $sub_res");
        $self->log("sbatch stderr $stderr");

        open(ERR, ">" . $description->stderr());
        print ERR $stderr;
        close(ERR);

        $stderr =~ s/\n/\\n/g;
        $self->respond({GT3_FAILURE_MESSAGE => $stderr });
    }

    return Globus::GRAM::Error::JOB_EXECUTION_FAILED();
}

sub poll
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $job_id = $description->jobid();
    my $state;
    my $exit_code;
    my $pid;
    my @out;

    $self->log("polling job $job_id");
    #open(MYOUTFILE, ">/tmp/my.log");
    #select MYOUTFILE; $| = 1;  # make unbuffered
    #print MYOUTFILE "JOB ID: $job_id\n";

    # Get job state line from the scontrol command output.
    local(*SCONTROL_IN, *SCONTROL_OUT, *SCONTROL_ERR);
    $pid = IPC::Open3::open3(\*SCONTROL_IN, \*SCONTROL_OUT, \*SCONTROL_ERR, "$scontrol", 'show', 'job', "$job_id");
    close(SCONTROL_IN);
    waitpid($pid, 0);

    # Get the exit code of the scontrol command.  For info search $CHILD_ERROR
    # in perlvar documentation or C documentation about waitpid.
    $exit_code = $? >> 8;

    # scontrol returns 1 on any error, see comment in scontrol.c code.
    if ($exit_code == 1) {
        local $/;
        $out = (grep(/slurm_load_jobs error:/, <SCONTROL_ERR>))[0];
        if (defined $out) {
            chomp ($out);
            # Assume the job was complited.
            if ($out =~ /slurm_load_jobs error: Invalid job id specified/) {
                $self->log("scontrol: $out");
    		return {JOB_STATE => Globus::GRAM::JobState::DONE};
            }
            # We lost connection to the slurm, tell ignore the poll by an
            # empty hash. 
            if ($out =~ /slurm_load_jobs error: Unable to contact slurm controller/) {
                $self->log("scontrol: Unable to contact slurm");
                return {};
            }
            # Unknown slurm error.  Just treat as local slurm problems, tell
            # to ignore the poll by an empty hash.
            $self->log("scontrol: unknown slurm error: $out");
            return {};
        }
        # probably non-scontrol error, just bypass it to the next section.
    }

    if ($exit_code != 0) {
        local $/;
        @out = <SCONTROL_ERR>;
        if (defined ((grep(/^open3:/, @$out))[0])) {
            $self->log("slurm.pm: @out");
        } else {
            # Should return some special error in this case???
            $self->log("slurm.pm: @out");
        }
        # We are failed to find out job state, so just tell to ignore
        # the poll by an empty hash.
        return {};
    }

    #{
    #    local $/;
	@out = <SCONTROL_OUT>;
    #    chomp(@out);
    #}

    $_ = (grep(/JobState/, @out))[0];
    $_ =~ s/^.*JobState=(\S+)\s.*$/$1/ unless not defined;
    #print MYOUTFILE "var info: $_\n";
    if (not defined or $_ =~ '') {
        # We are failed to find out job state, so just tell to ignore
        # the poll by an empty hash.
        $self->log("slurm.pm: @out");
        return {};
    }

    if (/PENDING|CONFIGURING/) {
        $state = Globus::GRAM::JobState::PENDING;
    } elsif (/RUNNING|COMPLETING/) {
        $state = Globus::GRAM::JobState::ACTIVE;
    } elsif (/FAILED|CANCELLED|NODE_FAIL/) {
        $state = Globus::GRAM::JobState::FAILED;
    } elsif (/COMPLETED|TIMEOUT/) {
        $state = Globus::GRAM::JobState::DONE;
        $self->nfssync( $description->stdout() )
            if $description->stdout() ne '';
        $self->nfssync( $description->stderr() )
            if $description->stderr() ne '';
    } elsif (/SUSPENDED/) {
        $state = Globus::GRAM::JobState::SUSPENDED;
    } else {
        # We got unknown job state, so just tell to ignore
        # the poll by an empty hash.
        $self->log("slurm.pm: @out");
        return {};
    }

    #print MYOUTFILE "we are almost finished: $state\n";
    return {JOB_STATE => $state};
}

sub cancel
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $job_id = $description->jobid();

    $self->log("cancel job $job_id");

    $self->fork_and_exec_cmd($scancel, "--signal=KILL", $job_id);

    if ($? == 0) {
        return { JOB_STATE => Globus::GRAM::JobState::FAILED }
    }

    return Globus::GRAM::Error::JOB_CANCEL_FAILED();
}

1;
